"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const repo_1 = __importDefault(require("../repo"));
const community_1 = require("../community/community");
const crypto_1 = require("../crypto");
const services_pb_1 = require("tupelo-messages/services/services_pb");
const node_1 = require("../node");
const chaintree_1 = require("../chaintree");
const cids_1 = __importDefault(require("cids"));
const tupelo_1 = __importDefault(require("../tupelo"));
const debug_1 = __importDefault(require("debug"));
const log = debug_1.default("legacy");
const Key = require("interface-datastore").Key;
const pull = require('pull-stream/pull');
class TupeloClient {
    constructor(opts) {
        this.repo = this._createRepo(opts);
        this.notaryGroup = opts.notaryGroup;
    }
    /*
    start starts the community and opens the repo
    */
    start() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.repo.init({});
            yield this.repo.open();
            yield this._startCommunity();
            return;
        });
    }
    // close closes the repo
    close() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.node !== undefined) {
                this.node.stop();
            }
            return this.repo.close();
        });
    }
    // register is deprecated but here to avoid breaking the API
    register() {
        return __awaiter(this, void 0, void 0, function* () {
            return;
            // return promiseAroundRpcCallback((clbk) => {
            //   var req = new services.RegisterWalletRequest();
            //   req.setCreds(this.walletCreds);
            //   this.rpc.register(req, clbk);
            // });
        });
    }
    generateKey() {
        return __awaiter(this, void 0, void 0, function* () {
            const ecdsaKey = yield crypto_1.EcdsaKey.generate();
            const keyAddr = yield ecdsaKey.keyAddr();
            const dbKey = new Key("/privatekeys/" + keyAddr);
            if (ecdsaKey.privateKey === undefined) {
                throw new Error("got undefined private key from key generation");
            }
            const resp = new services_pb_1.GenerateKeyResponse();
            resp.setKeyAddr(yield ecdsaKey.keyAddr());
            yield this.repo.put(dbKey, ecdsaKey.privateKey);
            return resp;
        });
    }
    listKeys() {
        return __awaiter(this, void 0, void 0, function* () {
            let resolve, reject;
            const p = new Promise((res, rej) => { resolve = res, reject = rej; });
            const resp = new services_pb_1.ListKeysResponse();
            pull(this.repo.query({
                prefix: "/privatekeys/"
            }), pull.collect((err, list) => __awaiter(this, void 0, void 0, function* () {
                if (err !== null) {
                    console.error('error in query: ', err);
                    reject(err);
                }
                let stringList = [];
                for (let i = 0; i < list.length; i++) {
                    let ecdsaKey = yield crypto_1.EcdsaKey.fromBytes(list[i].value);
                    stringList[i] = yield ecdsaKey.keyAddr();
                }
                resp.setKeyAddrsList(stringList);
                resolve(resp);
            })));
            return p;
        });
    }
    createChainTree(keyAddr) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.community === undefined) {
                throw new Error("community is undefined");
            }
            const resp = new services_pb_1.GenerateChainResponse();
            let key;
            try {
                key = yield this._getKey(keyAddr);
            }
            catch (e) {
                throw new Error("error getting key from addr: " + e.message);
            }
            const tree = yield chaintree_1.ChainTree.newEmptyTree(this.community.blockservice, key);
            const id = yield tree.id();
            if (id == undefined) {
                throw new Error("undefined id returned");
            }
            yield this.repo.put(new Key("/chaintrees/" + id), tree.tip.buffer);
            resp.setChainId(id);
            return resp;
        });
    }
    listChainIds() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.community === undefined) {
                throw new Error("community is undefined");
            }
            const resp = new services_pb_1.ListChainIdsResponse();
            let resolve, reject;
            const p = new Promise((res, rej) => { resolve = res, reject = rej; });
            pull(this.repo.query({
                prefix: "/chaintrees/"
            }), pull.collect((err, list) => __awaiter(this, void 0, void 0, function* () {
                if (err !== null) {
                    console.error('error in query: ', err);
                    reject(err);
                }
                let stringList = [];
                for (let i = 0; i < list.length; i++) {
                    let key = list[i].key.toString().slice("/chaintrees/".length);
                    stringList[i] = key;
                }
                resp.setChainIdsList(stringList);
                resolve(resp);
            })));
            return p;
        });
    }
    getTip(chainId) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.community === undefined) {
                throw new Error("community is undefined");
            }
            const resp = new services_pb_1.GetTipResponse();
            const tip = yield this.community.getTip(chainId);
            resp.setTip(tip.toString());
            return resp;
        });
    }
    playTransactions(chainId, keyAddr, transactions) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.community === undefined) {
                throw new Error("community is undefined");
            }
            const key = yield this._getKey(keyAddr);
            const treeTipBits = yield this.repo.get("/chaintrees/" + chainId);
            const treeTip = new cids_1.default(treeTipBits);
            const tree = new chaintree_1.ChainTree({
                store: this.community.blockservice,
                tip: treeTip,
                key: key,
            });
            const currState = yield tupelo_1.default.playTransactions(this.community.node.pubsub, this.community.group, tree, transactions);
            const resp = new services_pb_1.PlayTransactionsResponse();
            const sig = currState.getSignature();
            if (sig == undefined) {
                throw new Error("empty signature received from CurrState");
            }
            const tip = new cids_1.default(Buffer.from(sig.getNewTip_asU8()));
            resp.setTip(tip.toString());
            return resp;
        });
    }
    _getKey(addr) {
        return __awaiter(this, void 0, void 0, function* () {
            let bits = yield this.repo.get(new Key("/privatekeys/" + addr));
            return crypto_1.EcdsaKey.fromBytes(bits);
        });
    }
    _startCommunity() {
        return __awaiter(this, void 0, void 0, function* () {
            const node = yield node_1.p2p.createNode({ bootstrapAddresses: this.notaryGroup.getBootstrapAddressesList() });
            node.on('error', (err) => {
                console.error('p2p error: ', err);
            });
            node.once('peer:connect', () => __awaiter(this, void 0, void 0, function* () {
                log("peer connected");
            }));
            node.start(() => {
                log("node started");
            });
            const c = new community_1.Community(node, this.notaryGroup, this.repo.repo);
            this.community = c;
            return c.start();
        });
    }
    _createRepo(opts) {
        const repo = new repo_1.default('_DEFAULT', {
            lock: 'memory',
            storageBackends: {
                root: opts.keystore,
                blocks: opts.blockstore,
                keys: opts.keystore,
                datastore: opts.keystore,
            }
        });
        return repo;
    }
}
exports.TupeloClient = TupeloClient;
